--!parent
--!encoding=utf-8

--****************************************************************************
-- @file      LSW CORE LIB: CONSOLE
-- @author    Valentin Schmidt
-- @version   0.3
--****************************************************************************

property pTimers
property pPrompt
property pOpen

property pUseBuddyAPI
property pWinHandle

----------------------------------------
-- @constructor
----------------------------------------
on new (me)
	$ = me.$
	
  $.include(me.libdir&"quit.ls")

  me.pOpen = FALSE
  me.pTimers = [:]
  me.pPrompt = RETURN&\
  "----------------------------------------"&RETURN&\
  "  Welcome to the LSW Console Window!"&RETURN&\
  "  If you close the console in a window-less application, the application keeps running in the background."&RETURN&\
  "  To force such an application to quit, enter q + <RETURN> instead."&RETURN&\
  "----------------------------------------"

  me.pUseBuddyAPI = FALSE

  if me.pUseBuddyAPI then
    $.loadXtras(me.libdir, ["BudAPI.x32"], [])
    ok = baRegister("fluxus", 2351520)
  end if

  return me
end

----------------------------------------
--
----------------------------------------
on show (me)
  if me.pUseBuddyAPI then
    if voidP(me.pWinHandle) then
      _player.debugplaybackenabled = 1
      me.pWinHandle = baFindWindow( "" , "Message" )
      baRemoveSysItems( me.pWinHandle, 1, 0, 0 )
      put me.pPrompt
      me.pOpen = TRUE
    else if not me.pOpen then
      baSetWindowState(me.pWinHandle, "Normal")
      me.pOpen = TRUE
    end if
  else
    if not me.pOpen then
      _player.debugplaybackenabled = 1
      put me.pPrompt
      me.pOpen = TRUE
    end if
  end if
  return me
end

----------------------------------------
--
----------------------------------------
on hide (me)
  if me.pUseBuddyAPI then
    baSetWindowState(me.pWinHandle, "Hidden")
    me.pOpen = FALSE
  end if
end

----------------------------------------
--
----------------------------------------
on toggle (me)
  if me.pOpen then
    me.hide()
  else
    me.show()
  end if
end

----------------------------------------
--
----------------------------------------
on close (me)
  _player.debugplaybackenabled = 0
end

--**************************************
-- Console API
--**************************************

----------------------------------------
-- console.assert(assertion, obj1 [, obj2, ..., objN])
-- console.assert(assertion, msg [, subst1, ..., substN])
----------------------------------------
on assert (me, assertion)
  if assertion<>true then
    str = ""
    repeat with i = 3 to the paramCount
      put param(i)&" " after str
    end repeat
    delete the last char of str
    put str
  end if
end

----------------------------------------
-- console.clear()
----------------------------------------
on clear (me)
  --
end

----------------------------------------
-- console.count(label)
----------------------------------------
on count (me)
  --
end

----------------------------------------
-- console.dir(object)
----------------------------------------
on dir (me, obj)
  put RETURN&\
  "############################################################"&RETURN&\
  me._list(obj)&RETURN&\
  "############################################################"
end

----------------------------------------
-- console.error(obj1 [, obj2, ..., objN)
-- console.error(msg [, subst1, ..., substN)
----------------------------------------
on error (me, msg)
  str = RETURN&\
  "############################################################"&RETURN&\
  "ERROR: " & msg&RETURN&\
  "############################################################"
  put str
end

----------------------------------------
-- console.info(obj1 [, obj2, ..., objN)
-- console.info(msg [, subst1, ..., substN)
----------------------------------------
on info (me, msg)
  str = RETURN&\
  "############################################################"&RETURN&\
  "INFO: " & msg&RETURN&\
  "############################################################"
  put str
end

----------------------------------------
-- console.log(obj1 [, obj2, ..., objN)
-- console.log(msg [, subst1, ..., substN)
----------------------------------------
on log (me)
  str = ""
  repeat with i = 2 to the paramCount
    put param(i)&" " after str
  end repeat
  delete the last char of str
  put str
end

----------------------------------------
-- console.table(data [, columns])
----------------------------------------
on table (me, data, labels)
  if not listP(data) then return
  if count(data)=0 then return
  extra = 2
  
  rowCnt = count(data)
  
  -- normalize
  repeat with row = 1 to rowCnt
    if not listP(data[row]) then data[row]=[data[row]]
  end repeat
  
  -- find col widths
  w = []
  if listP(labels) then
    colCnt = count(labels)
    repeat with col = 1 to colCnt
      w[col] = length(labels[col])
    end repeat
  else
    colCnt = count(data[1])
    w[colCnt] = 0
  end if
  repeat with row = 1 to rowCnt
    repeat with col = 1 to colCnt
      w[col] = max(w[col], length(data[row][col]))
    end repeat
  end repeat
  
  str = RETURN
  if listP(labels) then
    put "     " after str
    repeat with col = 1 to colCnt
      put me._rpadd(labels[col],w[col]+extra) after str
    end repeat
    put RETURN after str
  end if
  repeat with row = 1 to rowCnt
    put me._rpadd(row, 5) after str
    repeat with col = 1 to colCnt
      put me._rpadd(data[row][col],w[col]+extra) after str
    end repeat
    put RETURN after str
  end repeat
  put str
end

----------------------------------------
-- console.time(timerName)
----------------------------------------
on time (me, timerName)
  me.pTimers[timerName] = the milliseconds
  me.log(timerName&": timer started")
end

----------------------------------------
-- console.timeEnd(timerName)
----------------------------------------
on timeEnd(me, timerName)
  me.log(timerName&": "&(the milliseconds-me.pTimers[timerName])&" ms")
  me.pTimers.deleteOne(timerName)
end

----------------------------------------
-- console.warn(obj1 [, obj2, ..., objN]);
-- console.warn(msg [, subst1, ..., substN]);
----------------------------------------
on warn (me, msg)
  str = RETURN&\
  "############################################################"&RETURN&\
  "WARN: " & msg&RETURN&\
  "############################################################"
  put str
end


--**************************************
-- PRIVATE
--**************************************

----------------------------------------
--
----------------------------------------
on _rpadd (me, str, len)
  str = string(str)
  padStr = ""
  repeat with i = 1 to len
    put " " after padStr
  end repeat
  return str & padStr.char[1..len-str.length]
end

----------------------------------------
--
----------------------------------------
on _list (me, obj, indent)
  -- config
  showHandlersFlag = 1

  indent = string(indent)
  --if indent="" then put obj
  t = ilk(obj)
  case (t) of
    #list:
      cnt = count(obj)
      str = "<LIST CNT="&cnt&">"
      repeat with i = 1 to cnt
        put RETURN&indent&i&": "&me._enc(obj[i], indent) after str -- "<"&ilk(obj[i])&">"
      end repeat
      return str

    #propList:
      cnt = count(obj)
      str = "<PROPLIST CNT="&cnt&">"
      repeat with i = 1 to cnt
        put RETURN&indent&me._enc(obj.getPropAt(i))&": "&me._enc(obj[i], indent) after str
      end repeat
      return str

    #instance:
      if (string(obj) starts "<Xtra ") then return indent&obj
      cnt = count(obj)
      name = string(obj)
      name = name.word[2..name.word.count-2]
      str = "<INSTANCE OF "&name&" CNT="&cnt&">"
      repeat with i = 1 to cnt
        propname = obj.getPropAt(i)
        v = value("obj."&propname) -- obj[i]
        put RETURN&indent&me._enc(propname)&": "&me._enc(v, indent) after str
      end repeat
      if showHandlersFlag then
        h = obj.handlers()
        h.sort()
        put RETURN&indent&"HANDLERS: "&me._enc(h, indent) after str
      else
        cnt = count(obj.handlers())
        put RETURN&indent&"HANDLERS <LIST CNT="&cnt&">" after str
      end if
      return str

    #script:
      cnt = count(obj)
      str = "<SCRIPT CNT="&cnt&">"
      repeat with i = 1 to cnt
        propname = obj.getPropAt(i)
        put RETURN&indent&me._enc(propname) after str
      end repeat
      if showHandlersFlag then
        h = obj.handlers()
        h.sort()
        put RETURN&indent&"HANDLERS: "&me._enc(h, indent) after str
      else
        cnt = count(obj.handlers())
        put RETURN&indent&"HANDLERS <LIST CNT="&cnt&">" after str
      end if
      return str

    otherwise:
      return indent&me._enc(obj)
  end case
end

----------------------------------------
--
----------------------------------------
on _enc (me, val, indent)
  case (ilk(val)) of
    #symbol:
      return "#"&val
    #string:
      return QUOTE&val&QUOTE
    #list, #propList, #instance:
      return me._list(val, indent&"    ")
    otherwise:
      return string(val)
  end case
end
